# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instrumentation for FastAPI
https://fastapi.tiangolo.com/
"""

from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple

try:
    import os
    import signal

    import fastapi
    import wrapt
    from fastapi import HTTPException
    from fastapi.exception_handlers import http_exception_handler
    from starlette.middleware import Middleware

    from instana.instrumentation.asgi import InstanaASGIMiddleware
    from instana.log import logger
    from instana.util.gunicorn import running_in_gunicorn
    from instana.util.traceutils import get_tracer_tuple

    from opentelemetry.semconv.trace import SpanAttributes

    if TYPE_CHECKING:
        from starlette.requests import Request
        from starlette.responses import Response

    if not (    # pragma: no cover
        hasattr(fastapi, "__version__")
        and (
            fastapi.__version__[0] > "0" or int(fastapi.__version__.split(".")[1]) >= 51
        )
    ):
        logger.debug(
            "Instana supports FastAPI package versions 0.51.0 and newer.  Skipping."
        )
        raise ImportError

    async def instana_exception_handler(
        request: "Request", exc: HTTPException
    ) -> "Response":
        """
        We capture FastAPI HTTPException, log the error and pass it on
        to the default exception handler.
        """
        try:
            _, span, _ = get_tracer_tuple()

            if span:
                if hasattr(exc, "detail") and 500 <= exc.status_code:
                    span.set_attribute("http.error", exc.detail)
                span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, exc.status_code)
        except Exception:
            logger.debug("FastAPI instana_exception_handler: ", exc_info=True)

        return await http_exception_handler(request, exc)

    @wrapt.patch_function_wrapper("fastapi.applications", "FastAPI.__init__")
    def init_with_instana(
        wrapped: Callable[..., fastapi.applications.FastAPI.__init__],
        instance: fastapi.applications.FastAPI,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> None:
        middleware = kwargs.get("middleware")
        if middleware is None:
            kwargs["middleware"] = [Middleware(InstanaASGIMiddleware)]
        elif isinstance(middleware, list):
            middleware.append(Middleware(InstanaASGIMiddleware))

        exception_handlers = kwargs.get("exception_handlers")
        if exception_handlers is None:
            kwargs["exception_handlers"] = dict()

        if isinstance(kwargs["exception_handlers"], dict):
            kwargs["exception_handlers"][HTTPException] = instana_exception_handler

        return wrapped(*args, **kwargs)

    logger.debug("Instrumenting FastAPI")

    # Reload GUnicorn when we are instrumenting an already running application
    if "INSTANA_MAGIC" in os.environ and running_in_gunicorn(): # pragma: no cover
        os.kill(os.getpid(), signal.SIGHUP)

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Base class for all the agent flavors
"""

import logging

import requests

from instana.log import logger


class BaseAgent(object):
    """Base class for all agent flavors"""

    client = None
    options = None

    def __init__(self):
        self.client = requests.Session()

    def update_log_level(self):
        """Uses the value in <self.log_level> to update the global logger"""
        if self.options is None or self.options.log_level not in [
            logging.DEBUG,
            logging.INFO,
            logging.WARN,
            logging.ERROR,
        ]:
            logger.warning("BaseAgent.update_log_level: Unknown log level set")
            return

        logger.setLevel(self.options.log_level)



# (c) Copyright IBM Corp. 2024


class UnsupportedFormatException(Exception):
    """UnsupportedFormatException should be used when the provided format
    value is unknown or disallowed by the :class:`InstanaTracer`.

    See :meth:`InstanaTracer.inject()` and :meth:`InstanaTracer.extract()`.
    """

    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import uvicorn

from tests.helpers import testenv

testenv["starlette_host"] = "127.0.0.1"
testenv["starlette_port"] = 10817
testenv["starlette_server"] = "http://" + testenv["starlette_host"] + ":" + str(testenv["starlette_port"])



def launch_starlette():
    from .app import starlette_server
    from instana.singletons import agent

    # Hack together a manual custom headers list; We'll use this in tests
    agent.options.extra_http_headers = [
        "X-Capture-This", 
        "X-Capture-That",
    ]

    uvicorn.run(
        starlette_server,
        host=testenv["starlette_host"],
        port=testenv["starlette_port"],
        log_level="critical",
    )



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import time
from typing import Generator, List

from celery import shared_task
import celery
import celery.app
import celery.contrib
import celery.contrib.testing
import celery.contrib.testing.worker
import pytest

from instana.singletons import tracer
from instana.span.span import InstanaSpan
from tests.helpers import get_first_span_by_filter

# TODO: Refactor to class based tests


@shared_task
def add(
    x: int,
    y: int,
) -> int:
    return x + y


@shared_task
def will_raise_error() -> None:
    raise Exception("This is a simulated error")


def filter_out_ping_tasks(
    spans: List[InstanaSpan],
) -> List[InstanaSpan]:
    filtered_spans = []
    for span in spans:
        is_ping_task = (
            span.n == "celery-worker" and span.data["celery"]["task"] == "celery.ping"
        )
        if not is_ping_task:
            filtered_spans.append(span)
    return filtered_spans


class TestCelery:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        yield

    def test_apply_async(
        self,
        celery_app: celery.app.base.Celery,
        celery_worker: celery.contrib.testing.worker.TestWorkController,
    ) -> None:
        with tracer.start_as_current_span("test"):
            _ = add.apply_async(args=(4, 5))

        # Wait for jobs to finish
        time.sleep(1)

        spans = filter_out_ping_tasks(self.recorder.queued_spans())
        assert len(spans) == 3

        def filter(span):
            return span.n == "sdk"

        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        def filter(span):
            return span.n == "celery-client"

        client_span = get_first_span_by_filter(spans, filter)
        assert client_span

        def filter(span):
            return span.n == "celery-worker"

        worker_span = get_first_span_by_filter(spans, filter)
        assert worker_span

        assert client_span.t == test_span.t
        assert client_span.t == worker_span.t
        assert client_span.p == test_span.s

        assert client_span.data["celery"]["task"] == "tests.frameworks.test_celery.add"
        assert client_span.data["celery"]["scheme"] == "redis"
        assert client_span.data["celery"]["host"] == "localhost"
        assert client_span.data["celery"]["port"] == "6379"
        assert client_span.data["celery"]["task_id"]
        assert not client_span.data["celery"]["error"]
        assert not client_span.ec

        assert worker_span.data["celery"]["task"] == "tests.frameworks.test_celery.add"
        assert worker_span.data["celery"]["scheme"] == "redis"
        assert worker_span.data["celery"]["host"] == "localhost"
        assert worker_span.data["celery"]["port"] == "6379"
        assert worker_span.data["celery"]["task_id"]
        assert not worker_span.data["celery"]["error"]
        assert not worker_span.data["celery"]["retry-reason"]
        assert not worker_span.ec

    def test_delay(
        self,
        celery_app: celery.app.base.Celery,
        celery_worker: celery.contrib.testing.worker.TestWorkController,
    ) -> None:
        with tracer.start_as_current_span("test"):
            _ = add.delay(4, 5)

        # Wait for jobs to finish
        time.sleep(0.5)

        spans = filter_out_ping_tasks(self.recorder.queued_spans())
        assert len(spans) == 3

        def filter(span):
            return span.n == "sdk"

        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        def filter(span):
            return span.n == "celery-client"

        client_span = get_first_span_by_filter(spans, filter)
        assert client_span

        def filter(span):
            return span.n == "celery-worker"

        worker_span = get_first_span_by_filter(spans, filter)
        assert worker_span

        assert client_span.t == test_span.t
        assert client_span.t == worker_span.t
        assert client_span.p == test_span.s

        assert client_span.data["celery"]["task"] == "tests.frameworks.test_celery.add"
        assert client_span.data["celery"]["scheme"] == "redis"
        assert client_span.data["celery"]["host"] == "localhost"
        assert client_span.data["celery"]["port"] == "6379"
        assert client_span.data["celery"]["task_id"]
        assert not client_span.data["celery"]["error"]
        assert not client_span.ec

        assert worker_span.data["celery"]["task"] == "tests.frameworks.test_celery.add"
        assert worker_span.data["celery"]["scheme"] == "redis"
        assert worker_span.data["celery"]["host"] == "localhost"
        assert worker_span.data["celery"]["port"] == "6379"
        assert worker_span.data["celery"]["task_id"]
        assert not worker_span.data["celery"]["error"]
        assert not worker_span.data["celery"]["retry-reason"]
        assert not worker_span.ec

    def test_send_task(
        self,
        celery_app: celery.app.base.Celery,
        celery_worker: celery.contrib.testing.worker.TestWorkController,
    ) -> None:
        with tracer.start_as_current_span("test"):
            _ = celery_app.send_task("tests.frameworks.test_celery.add", (1, 2))

        # Wait for jobs to finish
        time.sleep(0.5)

        spans = filter_out_ping_tasks(self.recorder.queued_spans())
        assert len(spans) == 3

        def filter(span):
            return span.n == "sdk"

        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        def filter(span):
            return span.n == "celery-client"

        client_span = get_first_span_by_filter(spans, filter)
        assert client_span

        def filter(span):
            return span.n == "celery-worker"

        worker_span = get_first_span_by_filter(spans, filter)
        assert worker_span

        assert client_span.t == test_span.t
        assert client_span.t == worker_span.t
        assert client_span.p == test_span.s

        assert client_span.data["celery"]["task"] == "tests.frameworks.test_celery.add"
        assert client_span.data["celery"]["scheme"] == "redis"
        assert client_span.data["celery"]["host"] == "localhost"
        assert client_span.data["celery"]["port"] == "6379"
        assert client_span.data["celery"]["task_id"]
        assert not client_span.data["celery"]["error"]
        assert not client_span.ec

        assert worker_span.data["celery"]["task"] == "tests.frameworks.test_celery.add"
        assert worker_span.data["celery"]["scheme"] == "redis"
        assert worker_span.data["celery"]["host"] == "localhost"
        assert worker_span.data["celery"]["port"] == "6379"
        assert worker_span.data["celery"]["task_id"]
        assert not worker_span.data["celery"]["error"]
        assert not worker_span.data["celery"]["retry-reason"]
        assert not worker_span.ec

    def test_error_reporting(
        self,
        celery_app: celery.app.base.Celery,
        celery_worker: celery.contrib.testing.worker.TestWorkController,
    ) -> None:
        with tracer.start_as_current_span("test"):
            _ = will_raise_error.apply_async()

        # Wait for jobs to finish
        time.sleep(4)

        spans = filter_out_ping_tasks(self.recorder.queued_spans())
        assert len(spans) == 4

        def filter(span):
            return span.n == "sdk"

        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        def filter(span):
            return span.n == "celery-client"

        client_span = get_first_span_by_filter(spans, filter)
        assert client_span

        def filter(span):
            return span.n == "log"

        log_span = get_first_span_by_filter(spans, filter)
        assert log_span

        def filter(span):
            return span.n == "celery-worker"

        worker_span = get_first_span_by_filter(spans, filter)
        assert worker_span

        assert client_span.t == test_span.t
        assert client_span.t == worker_span.t
        assert client_span.t == log_span.t

        assert client_span.p == test_span.s
        assert worker_span.p == client_span.s
        assert log_span.p == worker_span.s

        assert (
            client_span.data["celery"]["task"]
            == "tests.frameworks.test_celery.will_raise_error"
        )
        assert client_span.data["celery"]["scheme"] == "redis"
        assert client_span.data["celery"]["host"] == "localhost"
        assert client_span.data["celery"]["port"] == "6379"
        assert client_span.data["celery"]["task_id"]
        assert not client_span.data["celery"]["error"]
        assert not client_span.ec

        assert (
            worker_span.data["celery"]["task"]
            == "tests.frameworks.test_celery.will_raise_error"
        )
        assert worker_span.data["celery"]["scheme"] == "redis"
        assert worker_span.data["celery"]["host"] == "localhost"
        assert worker_span.data["celery"]["port"] == "6379"
        assert worker_span.data["celery"]["task_id"]
        assert worker_span.data["celery"]["error"] == "This is a simulated error"
        assert not worker_span.data["celery"]["retry-reason"]
        assert worker_span.ec == 1



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import sys
from .app import aiohttp_server as server
from ..utils import launch_background_thread

APP_THREAD = None

if not any((os.environ.get('GEVENT_STARLETTE_TEST'),
            os.environ.get('CASSANDRA_TEST'),
            sys.version_info < (3, 5, 3))):
    APP_THREAD = launch_background_thread(server, "AIOHTTP")



# (c) Copyright IBM Corp. 2024

import time


def timing_tween_factory(handler, registry):
    def timing_tween(request):
        start = time.time()
        try:
            response = handler(request)
        finally:
            end = time.time()
            print(f"The request took {end - start} seconds")
        return response

    return timing_tween



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

import os
import pytest

testenv = {}

"""
Cassandra Environment
"""
testenv["cassandra_host"] = os.environ.get("CASSANDRA_HOST", "127.0.0.1")
testenv["cassandra_username"] = os.environ.get("CASSANDRA_USERNAME", "Administrator")
testenv["cassandra_password"] = os.environ.get("CASSANDRA_PASSWORD", "password")

"""
CouchDB Environment
"""
testenv["couchdb_host"] = os.environ.get("COUCHDB_HOST", "127.0.0.1")
testenv["couchdb_username"] = os.environ.get("COUCHDB_USERNAME", "Administrator")
testenv["couchdb_password"] = os.environ.get("COUCHDB_PASSWORD", "password")

"""
MySQL Environment
"""
if "MYSQL_HOST" in os.environ:
    testenv["mysql_host"] = os.environ["MYSQL_HOST"]
else:
    testenv["mysql_host"] = "127.0.0.1"

testenv["mysql_port"] = int(os.environ.get("MYSQL_PORT", "3306"))
testenv["mysql_db"] = os.environ.get("MYSQL_DATABASE", "instana_test_db")
testenv["mysql_user"] = os.environ.get("MYSQL_USER", "root")
testenv["mysql_pw"] = os.environ.get("MYSQL_ROOT_PASSWORD", "passw0rd")

"""
PostgreSQL Environment
"""
testenv["postgresql_host"] = os.environ.get("POSTGRES_HOST", "127.0.0.1")
testenv["postgresql_port"] = int(os.environ.get("POSTGRES_PORT", "5432"))
testenv["postgresql_db"] = os.environ.get("POSTGRES_DB", "instana_test_db")
testenv["postgresql_user"] = os.environ.get("POSTGRES_USER", "root")
testenv["postgresql_pw"] = os.environ.get("POSTGRES_PW", "passw0rd")

"""
Redis Environment
"""
testenv["redis_host"] = os.environ.get("REDIS_HOST", "127.0.0.1")
testenv["redis_db"] = os.environ.get("REDIS_DB", 0)

"""
MongoDB Environment
"""
testenv["mongodb_host"] = os.environ.get("MONGO_HOST", "127.0.0.1")
testenv["mongodb_port"] = os.environ.get("MONGO_PORT", "27017")
testenv["mongodb_user"] = os.environ.get("MONGO_USER", None)
testenv["mongodb_pw"] = os.environ.get("MONGO_PW", None)


def drop_log_spans_from_list(spans):
    """
    Log spans may occur randomly in test runs because of various intentional errors (for testing). This
    helper method will remove all of the log spans from <spans> and return the remaining list.  Helpful
    for those tests where we are not testing log spans - where log spans are just noise.
    @param spans: the list of spans to filter
    @return: a filtered list of spans
    """
    new_list = []
    for span in spans:
        if span.n != "log":
            new_list.append(span)
    return new_list


def fail_with_message_and_span_dump(msg, spans):
    """
    Helper method to fail a test when the number of spans isn't what was expected.  This helper
    will print <msg> and dump the list of spans in <spans>.

    @param msg: Descriptive message to print with the failure
    @param spans: the list of spans to dump
    @return: None
    """
    span_count = len(spans)
    span_dump = "\nDumping all collected spans (%d) -->\n" % span_count
    if span_count > 0:
        for span in spans:
            span.stack = "<snipped>"
            span_dump += repr(span) + "\n"
    pytest.fail(msg + span_dump, True)


def get_first_span_by_name(spans, name):
    """
    Get the first span in <spans> that has a span.n value of <name>
    @param spans: the list of spans to search
    @param name: the name to search for
    @return: Span or None if nothing found
    """
    for span in spans:
        if span.n == name:
            return span
    return None


def get_first_span_by_filter(spans, filter):
    """
    Get the first span in <spans> that matches <filter>

    Example:
    filter = lambda span: span.n == "tornado-server" and span.data["http"]["status"] == 301
    tornado_301_span = get_first_span_by_filter(spans, filter)

    @param spans: the list of spans to search
    @param filter: the filter to search by
    @return: Span or None if nothing matched
    """
    for span in spans:
        if filter(span) is True:
            return span
    return None


def get_spans_by_filter(spans, filter):
    """
    Get all spans in <spans> that matches <filter>

    Example:
    filter = lambda span: span.n == "tornado-server" and span.data["http"]["status"] == 301
    tornado_301_spans = get_spans_by_filter(spans, filter)

    @param spans: the list of spans to search
    @param filter: the filter to search by
    @return: list of spans
    """
    results = []
    for span in spans:
        if filter(span) is True:
            results.append(span)
    return results


def launch_traced_request(url):
    import requests
    from instana.log import logger
    from instana.singletons import tracer

    logger.warn(
        "Launching request with a root SDK span name of 'launch_traced_request'"
    )

    with tracer.start_as_current_span("launch_traced_request"):
        response = requests.get(url)

    return response






# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018


from typing import Any, Callable, Dict, Tuple
import wrapt

from instana.log import logger
from instana.span.span import InstanaSpan
from instana.util.traceutils import get_tracer_tuple, tracing_is_off

try:
    import redis

    EXCLUDED_PARENT_SPANS = ["redis", "celery-client", "celery-worker"]

    def collect_attributes(
        span: InstanaSpan,
        instance: redis.client.Redis,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> None:
        try:
            ckw = instance.connection_pool.connection_kwargs

            span.set_attribute("driver", "redis-py")

            host = ckw.get("host", None)
            port = ckw.get("port", "6379")
            db = ckw.get("db", None)

            if host:
                url = f"redis://{host}:{port}"
                if db is not None:
                    url = f"{url}/{db}"
                span.set_attribute("connection", url)
        except Exception:
            logger.debug("redis.collect_attributes non-fatal error", exc_info=True)

    def execute_command_with_instana(
        wrapped: Callable[..., object],
        instance: redis.client.Redis,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        tracer, parent_span, operation_name = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        # If we're not tracing, just return
        if tracing_is_off() or (operation_name in EXCLUDED_PARENT_SPANS):
            return wrapped(*args, **kwargs)

        with tracer.start_as_current_span("redis", span_context=parent_context) as span:
            try:
                collect_attributes(span, instance, args, kwargs)
                if len(args) > 0:
                    span.set_attribute("command", args[0])

                rv = wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
                raise
            else:
                return rv

    def execute_with_instana(
        wrapped: Callable[..., object],
        instance: redis.client.Redis,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        tracer, parent_span, operation_name = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        # If we're not tracing, just return
        if tracing_is_off() or (operation_name in EXCLUDED_PARENT_SPANS):
            return wrapped(*args, **kwargs)

        with tracer.start_as_current_span("redis", span_context=parent_context) as span:
            try:
                collect_attributes(span, instance, args, kwargs)
                span.set_attribute("command", "PIPELINE")

                pipe_cmds = []
                for e in instance.command_stack:
                    pipe_cmds.append(e[0][0])
                span.set_attribute("subCommands", pipe_cmds)
            except Exception as e:
                # If anything breaks during K/V collection, just log a debug message
                logger.debug("Error collecting pipeline commands", exc_info=True)

            try:
                rv = wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    if redis.VERSION < (3, 0, 0):
        wrapt.wrap_function_wrapper(
            "redis.client", "BasePipeline.execute", execute_with_instana
        )
        wrapt.wrap_function_wrapper(
            "redis.client", "StrictRedis.execute_command", execute_command_with_instana
        )
    else:
        wrapt.wrap_function_wrapper(
            "redis.client", "Pipeline.execute", execute_with_instana
        )
        wrapt.wrap_function_wrapper(
            "redis.client", "Redis.execute_command", execute_command_with_instana
        )

        logger.debug("Instrumenting redis")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import time
import asyncio
import pytest
from typing import Generator

import tornado
from tornado.httpclient import AsyncHTTPClient
from instana.singletons import tracer, agent
from instana.span.span import get_current_span

from instana.util.ids import hex_id
import tests.apps.tornado_server
from tests.helpers import testenv, get_first_span_by_name, get_first_span_by_filter

class TestTornadoClient:

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """ Clear all spans before a test run """
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # New event loop for every test
        # self.loop = tornado.ioloop.IOLoop.current()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        self.http_client = AsyncHTTPClient()
        yield
        self.http_client.close()

    def test_get(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                return await self.http_client.fetch(testenv["tornado_server"] + "/")

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "tornado-server")
        client_span = get_first_span_by_name(spans, "tornado-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert client_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == server_span.data["http"]["url"]
        assert not server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "GET"
        # assert server_span.stack
        # assert type(server_span.stack) is list
        # assert len(server_span.stack) > 1

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == client_span.data["http"]["url"]
        assert client_span.data["http"]["method"] == "GET"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_post(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                return await self.http_client.fetch(testenv["tornado_server"] + "/", method="POST", body='asdf')

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "tornado-server")
        client_span = get_first_span_by_name(spans, "tornado-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert client_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == server_span.data["http"]["url"]
        assert not server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "POST"

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == client_span.data["http"]["url"]
        assert client_span.data["http"]["method"] == "POST"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_get_301(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                return await self.http_client.fetch(testenv["tornado_server"] + "/301")

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        assert len(spans) == 5

        server301_span = spans[0]
        server_span = spans[1]
        client_span = spans[2]
        client301_span = spans[3]
        test_span = spans[4]

        filter = lambda span: span.n == "tornado-server" and span.data["http"]["status"] == 301
        server301_span = get_first_span_by_filter(spans, filter)
        filter = lambda span: span.n == "tornado-server" and span.data["http"]["status"] == 200
        server_span = get_first_span_by_filter(spans, filter)
        filter = lambda span: span.n == "tornado-client" and span.data["http"]["url"] == testenv["tornado_server"] + "/"
        client_span = get_first_span_by_filter(spans, filter)
        filter = lambda span: span.n == "tornado-client" and span.data["http"]["url"] == testenv["tornado_server"] + "/301"
        client301_span = get_first_span_by_filter(spans, filter)
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == client301_span.t
        assert traceId == server301_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert server301_span.p == client301_span.s
        assert client_span.p == test_span.s
        assert client301_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == server_span.data["http"]["url"]
        assert not server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "GET"

        assert server301_span.n == "tornado-server"
        assert server301_span.data["http"]["status"] == 301
        assert testenv["tornado_server"] + "/301" == server301_span.data["http"]["url"]
        assert not server301_span.data["http"]["params"]
        assert server301_span.data["http"]["method"] == "GET"

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == client_span.data["http"]["url"]
        assert client_span.data["http"]["method"] == "GET"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert client301_span.n == "tornado-client"
        assert client301_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/301" == client301_span.data["http"]["url"]
        assert client301_span.data["http"]["method"] == "GET"
        assert client301_span.stack
        assert type(client301_span.stack) is list
        assert len(client301_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_get_405(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                try:
                    return await self.http_client.fetch(testenv["tornado_server"] + "/405")
                except tornado.httpclient.HTTPClientError as e:
                    return e.response

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "tornado-server")
        client_span = get_first_span_by_name(spans, "tornado-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert client_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert client_span.ec == 1
        assert not server_span.ec

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 405
        assert testenv["tornado_server"] + "/405" == server_span.data["http"]["url"]
        assert not server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "GET"

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 405
        assert testenv["tornado_server"] + "/405" == client_span.data["http"]["url"]
        assert client_span.data["http"]["method"] == "GET"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_get_500(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                try:
                    return await self.http_client.fetch(testenv["tornado_server"] + "/500")
                except tornado.httpclient.HTTPClientError as e:
                    return e.response

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "tornado-server")
        client_span = get_first_span_by_name(spans, "tornado-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert client_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert client_span.ec == 1
        assert server_span.ec == 1

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 500
        assert testenv["tornado_server"] + "/500" == server_span.data["http"]["url"]
        assert not server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "GET"

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 500
        assert testenv["tornado_server"] + "/500" == client_span.data["http"]["url"]
        assert client_span.data["http"]["method"] == "GET"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_get_504(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                try:
                    return await self.http_client.fetch(testenv["tornado_server"] + "/504")
                except tornado.httpclient.HTTPClientError as e:
                    return e.response

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "tornado-server")
        client_span = get_first_span_by_name(spans, "tornado-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert client_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert client_span.ec == 1
        assert server_span.ec == 1

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 504
        assert testenv["tornado_server"] + "/504" == server_span.data["http"]["url"]
        assert not server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "GET"

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 504
        assert testenv["tornado_server"] + "/504" == client_span.data["http"]["url"]
        assert client_span.data["http"]["method"] == "GET"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_get_with_params_to_scrub(self) -> None:
        async def test():
            with tracer.start_as_current_span("test"):
                return await self.http_client.fetch(testenv["tornado_server"] + "/?secret=yeah")

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "tornado-server")
        client_span = get_first_span_by_name(spans, "tornado-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert client_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == server_span.data["http"]["url"]
        assert 'secret=<redacted>' == server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "GET"

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == client_span.data["http"]["url"]
        assert 'secret=<redacted>' == client_span.data["http"]["params"]
        assert client_span.data["http"]["method"] == "GET"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_request_header_capture(self) -> None:
        # Hack together a manual custom headers list
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        request_headers = {
            "X-Capture-This": "this",
            "X-Capture-That": "that",
        }

        async def test():
            with tracer.start_as_current_span("test"):
                return await self.http_client.fetch(testenv["tornado_server"] + "/", headers=request_headers)

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "tornado-server")
        client_span = get_first_span_by_name(spans, "tornado-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert client_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == server_span.data["http"]["url"]
        assert not server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "GET"

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/" == client_span.data["http"]["url"]
        assert client_span.data["http"]["method"] == "GET"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

        assert "X-Capture-This" in client_span.data["http"]["header"]
        assert client_span.data["http"]["header"]["X-Capture-This"] == "this"
        assert "X-Capture-That" in client_span.data["http"]["header"]
        assert client_span.data["http"]["header"]["X-Capture-That"] == "that"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_response_header_capture(self) -> None:
        # Hack together a manual custom headers list
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        async def test():
            with tracer.start_as_current_span("test"):
                return await self.http_client.fetch(testenv["tornado_server"] + "/response_headers")

        response = tornado.ioloop.IOLoop.current().run_sync(test)
        assert isinstance(response, tornado.httpclient.HTTPResponse)

        time.sleep(0.5)
        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "tornado-server")
        client_span = get_first_span_by_name(spans, "tornado-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert not get_current_span().is_recording()

        # Same traceId
        traceId = test_span.t
        assert traceId == client_span.t
        assert traceId == server_span.t

        # Parent relationships
        assert client_span.p == test_span.s
        assert server_span.p == client_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        assert server_span.n == "tornado-server"
        assert server_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/response_headers" == server_span.data["http"]["url"]
        assert not server_span.data["http"]["params"]
        assert server_span.data["http"]["method"] == "GET"

        assert client_span.n == "tornado-client"
        assert client_span.data["http"]["status"] == 200
        assert testenv["tornado_server"] + "/response_headers" == client_span.data["http"]["url"]
        assert client_span.data["http"]["method"] == "GET"
        assert client_span.stack
        assert type(client_span.stack) is list
        assert len(client_span.stack) > 1

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(server_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == '1'
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

        assert "X-Capture-This-Too" in client_span.data["http"]["header"]
        assert client_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in client_span.data["http"]["header"]
        assert client_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

import threading
import time
from typing import Generator, Optional

import mock
import pika
import pika.adapters.blocking_connection
import pika.channel
import pika.spec
import pytest
from opentelemetry.trace.span import format_span_id

from instana.singletons import agent, tracer
from instana.util.ids import hex_id


class _TestPika:
    @staticmethod
    @mock.patch("pika.connection.Connection")
    def _create_connection(connection_class_mock=None) -> object:
        return connection_class_mock()

    def _create_obj(self) -> NotImplementedError:
        raise NotImplementedError()

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        self.connection = self._create_connection()
        self._on_openok_callback = mock.Mock()
        self.obj = self._create_obj()
        yield
        # teardown
        del self.connection
        del self._on_openok_callback
        del self.obj
        # Ensure that allow_exit_as_root has the default value
        agent.options.allow_exit_as_root = False


class TestPikaBlockingChannel(_TestPika):
    @mock.patch("pika.channel.Channel", spec=pika.channel.Channel)
    def _create_obj(
        self, channel_impl: mock.MagicMock
    ) -> pika.adapters.blocking_connection.BlockingChannel:
        self.impl = channel_impl()
        self.impl.channel_number = 1

        return pika.adapters.blocking_connection.BlockingChannel(
            self.impl, self.connection
        )

    def _generate_delivery(
        self, consumer_tag: str, properties: pika.BasicProperties, body: str
    ) -> None:
        from pika.adapters.blocking_connection import _ConsumerDeliveryEvt

        # Wait until queue consumer is initialized
        while self.obj._queue_consumer_generator is None:
            time.sleep(0.25)

        method = pika.spec.Basic.Deliver(consumer_tag=consumer_tag)
        self.obj._on_consumer_generator_event(
            _ConsumerDeliveryEvt(method, properties, body)
        )

    def test_consume(self) -> None:
        consumed_deliveries = []

        def __consume() -> None:
            for delivery in self.obj.consume("test.queue", inactivity_timeout=3.0):
                # Skip deliveries generated due to inactivity
                if delivery is not None and any(delivery):
                    consumed_deliveries.append(delivery)

                break

        consumer_tag = "test.consumer"

        self.impl.basic_consume.return_value = consumer_tag
        self.impl._generate_consumer_tag.return_value = consumer_tag
        self.impl._consumers = {}

        t = threading.Thread(target=__consume)
        t.start()

        self._generate_delivery(consumer_tag, pika.BasicProperties(), "Hello!")

        t.join(timeout=5.0)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # A new span has been started
        assert rabbitmq_span.t
        assert not rabbitmq_span.p
        assert rabbitmq_span.s

        # Error logging
        assert not rabbitmq_span.ec

        # Span tags
        assert not rabbitmq_span.data["rabbitmq"]["exchange"]
        assert rabbitmq_span.data["rabbitmq"]["sort"] == "consume"
        assert rabbitmq_span.data["rabbitmq"]["address"]
        assert rabbitmq_span.data["rabbitmq"]["queue"] == "test.queue"
        assert rabbitmq_span.stack
        assert isinstance(rabbitmq_span.stack, list)
        assert len(rabbitmq_span.stack) > 0

        assert len(consumed_deliveries) == 1

    def test_consume_with_trace_context(self) -> None:
        consumed_deliveries = []

        def __consume():
            for delivery in self.obj.consume("test.queue", inactivity_timeout=3.0):
                # Skip deliveries generated due to inactivity
                if delivery is not None and any(delivery):
                    consumed_deliveries.append(delivery)
                break

        consumer_tag = "test.consumer"

        self.impl.basic_consume.return_value = consumer_tag
        self.impl._generate_consumer_tag.return_value = consumer_tag
        self.impl._consumers = {}

        t = threading.Thread(target=__consume)
        t.start()

        instana_headers = {
            "X-INSTANA-T": "0000000000000001",
            "X-INSTANA-S": "0000000000000002",
            "X-INSTANA-L": "1",
        }
        self._generate_delivery(
            consumer_tag,
            pika.BasicProperties(headers=instana_headers),
            "Hello!",
        )

        t.join(timeout=5.0)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # Trace context propagation
        assert rabbitmq_span.t == int(instana_headers["X-INSTANA-T"])
        assert rabbitmq_span.p == int(instana_headers["X-INSTANA-S"])

        # A new span has been started
        assert rabbitmq_span.s
        assert rabbitmq_span.p != rabbitmq_span.s

    def test_consume_with_not_GeneratorType(self, mocker) -> None:
        mocker.patch(
            "instana.instrumentation.pika.isinstance",
            return_value=False,
        )

        consumed_deliveries = []

        def __consume() -> None:
            for delivery in self.obj.consume("test.queue", inactivity_timeout=3.0):
                # Skip deliveries generated due to inactivity
                if delivery is not None and any(delivery):
                    consumed_deliveries.append(delivery)

                break

        consumer_tag = "test.consumer"

        self.impl.basic_consume.return_value = consumer_tag
        self.impl._generate_consumer_tag.return_value = consumer_tag
        self.impl._consumers = {}

        t = threading.Thread(target=__consume)
        t.start()

        self._generate_delivery(consumer_tag, pika.BasicProperties(), "Hello!")

        t.join(timeout=5.0)

        spans = self.recorder.queued_spans()
        assert len(spans) == 0

    def test_consume_with_any_yielded(self, mocker) -> None:
        mocker.patch(
            "instana.instrumentation.pika.any",
            return_value=False,
        )

        consumed_deliveries = []

        def __consume() -> None:
            for delivery in self.obj.consume("test.queue", inactivity_timeout=3.0):
                # Skip deliveries generated due to inactivity
                if delivery is not None and any(delivery):
                    consumed_deliveries.append(delivery)

                break

        consumer_tag = "test.consumer"

        self.impl.basic_consume.return_value = consumer_tag
        self.impl._generate_consumer_tag.return_value = consumer_tag
        self.impl._consumers = {}

        t = threading.Thread(target=__consume)
        t.start()

        self._generate_delivery(consumer_tag, pika.BasicProperties(), "Hello!")

        t.join(timeout=5.0)

        spans = self.recorder.queued_spans()
        assert len(spans) == 0


class TestPikaBlockingChannelBlockingConnection(_TestPika):
    @mock.patch("pika.adapters.blocking_connection.BlockingConnection", autospec=True)
    def _create_connection(self, connection: Optional[mock.MagicMock] = None) -> object:
        connection._impl = mock.create_autospec(pika.connection.Connection)
        connection._impl.params = pika.connection.Parameters()
        return connection

    @mock.patch("pika.channel.Channel", spec=pika.channel.Channel)
    def _create_obj(
        self, channel_impl: mock.MagicMock
    ) -> pika.adapters.blocking_connection.BlockingChannel:
        self.impl = channel_impl()
        self.impl.channel_number = 1

        return pika.adapters.blocking_connection.BlockingChannel(
            self.impl, self.connection
        )

    def _generate_delivery(
        self,
        method: pika.spec.Basic.Deliver,
        properties: pika.BasicProperties,
        body: str,
    ) -> None:
        from pika.adapters.blocking_connection import _ConsumerDeliveryEvt

        evt = _ConsumerDeliveryEvt(method, properties, body)
        self.obj._add_pending_event(evt)
        self.obj._dispatch_events()

    def test_basic_consume(self) -> None:
        consumer_tag = "test.consumer"

        self.impl.basic_consume.return_value = consumer_tag
        self.impl._generate_consumer_tag.return_value = consumer_tag

        cb = mock.Mock()

        self.obj.basic_consume(queue="test.queue", on_message_callback=cb)

        body = "Hello!"
        properties = pika.BasicProperties()
        method = pika.spec.Basic.Deliver(consumer_tag)
        self._generate_delivery(method, properties, body)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # A new span has been started
        assert rabbitmq_span.t
        assert not rabbitmq_span.p
        assert rabbitmq_span.s

        # Error logging
        assert not rabbitmq_span.ec

        # Span tags
        assert not rabbitmq_span.data["rabbitmq"]["exchange"]
        assert rabbitmq_span.data["rabbitmq"]["sort"] == "consume"
        assert rabbitmq_span.data["rabbitmq"]["address"]
        assert rabbitmq_span.data["rabbitmq"]["queue"] == "test.queue"
        assert rabbitmq_span.stack
        assert isinstance(rabbitmq_span.stack, list)
        assert len(rabbitmq_span.stack) > 0

        cb.assert_called_once_with(self.obj, method, properties, body)

    def test_basic_consume_with_trace_context(self):
        consumer_tag = "test.consumer"

        self.impl.basic_consume.return_value = consumer_tag
        self.impl._generate_consumer_tag.return_value = consumer_tag

        cb = mock.Mock()

        self.obj.basic_consume(queue="test.queue", on_message_callback=cb)

        body = "Hello!"
        instana_headers = {
            "X-INSTANA-T": "0000000000000001",
            "X-INSTANA-S": "0000000000000002",
            "X-INSTANA-L": "1",
        }
        properties = pika.BasicProperties(headers=instana_headers)
        method = pika.spec.Basic.Deliver(consumer_tag)
        self._generate_delivery(method, properties, body)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # Trace context propagation
        assert rabbitmq_span.t == int(instana_headers["X-INSTANA-T"])
        assert rabbitmq_span.p == int(instana_headers["X-INSTANA-S"])

        # A new span has been started
        assert rabbitmq_span.s
        assert rabbitmq_span.p != rabbitmq_span.s


class TestPikaChannel(_TestPika):
    def _create_obj(self) -> pika.channel.Channel:
        return pika.channel.Channel(self.connection, 1, self._on_openok_callback)

    @mock.patch("pika.spec.Basic.Publish")
    @mock.patch("pika.channel.Channel._send_method")
    def test_basic_publish(self, send_method, _unused) -> None:
        self.obj._set_state(self.obj.OPEN)

        with tracer.start_as_current_span("testing"):
            self.obj.basic_publish("test.exchange", "test.queue", "Hello!")

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        rabbitmq_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == rabbitmq_span.t

        # Parent relationships
        assert rabbitmq_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rabbitmq_span.ec

        # Span tags
        assert rabbitmq_span.data["rabbitmq"]["exchange"] == "test.exchange"
        assert rabbitmq_span.data["rabbitmq"]["sort"] == "publish"
        assert rabbitmq_span.data["rabbitmq"]["address"]
        assert rabbitmq_span.data["rabbitmq"]["key"] == "test.queue"
        assert rabbitmq_span.stack
        assert isinstance(rabbitmq_span.stack, list)
        assert len(rabbitmq_span.stack) > 0

        send_method.assert_called_once_with(
            pika.spec.Basic.Publish(exchange="test.exchange", routing_key="test.queue"),
            (
                pika.spec.BasicProperties(
                    headers={
                        "X-INSTANA-T": format_span_id(rabbitmq_span.t),
                        "X-INSTANA-S": format_span_id(rabbitmq_span.s),
                        "X-INSTANA-L": "1",
                        "Server-Timing": f"intid;desc={hex_id(rabbitmq_span.t)}",
                    }
                ),
                b"Hello!",
            ),
        )

    @mock.patch("pika.spec.Basic.Publish")
    @mock.patch("pika.channel.Channel._send_method")
    def test_basic_publish_as_root_exit_span(self, send_method, _unused) -> None:
        agent.options.allow_exit_as_root = True
        self.obj._set_state(self.obj.OPEN)
        self.obj.basic_publish("test.exchange", "test.queue", "Hello!")

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # Parent relationships
        assert not rabbitmq_span.p

        # Error logging
        assert not rabbitmq_span.ec

        # Span tags
        assert rabbitmq_span.data["rabbitmq"]["exchange"] == "test.exchange"
        assert rabbitmq_span.data["rabbitmq"]["sort"] == "publish"
        assert rabbitmq_span.data["rabbitmq"]["address"]
        assert rabbitmq_span.data["rabbitmq"]["key"] == "test.queue"
        assert rabbitmq_span.stack
        assert isinstance(rabbitmq_span.stack, list)
        assert len(rabbitmq_span.stack) > 0

        send_method.assert_called_once_with(
            pika.spec.Basic.Publish(exchange="test.exchange", routing_key="test.queue"),
            (
                pika.spec.BasicProperties(
                    headers={
                        "X-INSTANA-T": format_span_id(rabbitmq_span.t),
                        "X-INSTANA-S": format_span_id(rabbitmq_span.s),
                        "X-INSTANA-L": "1",
                        "Server-Timing": f"intid;desc={hex_id(rabbitmq_span.t)}",
                    }
                ),
                b"Hello!",
            ),
        )

    @mock.patch("pika.spec.Basic.Publish")
    @mock.patch("pika.channel.Channel._send_method")
    def test_basic_publish_with_headers(self, send_method, _unused) -> None:
        self.obj._set_state(self.obj.OPEN)

        with tracer.start_as_current_span("testing"):
            self.obj.basic_publish(
                "test.exchange",
                "test.queue",
                "Hello!",
                pika.BasicProperties(headers={"X-Custom-1": "test"}),
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        rabbitmq_span = spans[0]

        send_method.assert_called_once_with(
            pika.spec.Basic.Publish(exchange="test.exchange", routing_key="test.queue"),
            (
                pika.spec.BasicProperties(
                    headers={
                        "X-Custom-1": "test",
                        "X-INSTANA-T": format_span_id(rabbitmq_span.t),
                        "X-INSTANA-S": format_span_id(rabbitmq_span.s),
                        "X-INSTANA-L": "1",
                        "Server-Timing": f"intid;desc={hex_id(rabbitmq_span.t)}",
                    }
                ),
                b"Hello!",
            ),
        )

    @mock.patch("pika.spec.Basic.Publish")
    @mock.patch("pika.channel.Channel._send_method")
    def test_basic_publish_tracing_off(self, send_method, _unused, mocker) -> None:
        mocker.patch(
            "instana.instrumentation.pika.tracing_is_off",
            return_value=True,
        )

        self.obj._set_state(self.obj.OPEN)

        with tracer.start_as_current_span("testing"):
            self.obj.basic_publish("test.exchange", "test.queue", "Hello!")

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        # Span names are not "rabbitmq"
        for span in spans:
            assert span.n != "rabbitmq"

    @mock.patch("pika.spec.Basic.Get")
    def test_basic_get(self, _unused) -> None:
        self.obj._set_state(self.obj.OPEN)

        body = "Hello!"
        properties = pika.BasicProperties()

        method_frame = pika.frame.Method(1, pika.spec.Basic.GetOk)
        header_frame = pika.frame.Header(1, len(body), properties)

        cb = mock.Mock()

        self.obj.basic_get("test.queue", cb)
        self.obj._on_getok(method_frame, header_frame, body)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # A new span has been started
        assert rabbitmq_span.t
        assert not rabbitmq_span.p
        assert rabbitmq_span.s

        # Error logging
        assert not rabbitmq_span.ec

        # Span tags
        assert not rabbitmq_span.data["rabbitmq"]["exchange"]
        assert rabbitmq_span.data["rabbitmq"]["sort"] == "consume"
        assert rabbitmq_span.data["rabbitmq"]["address"]
        assert rabbitmq_span.data["rabbitmq"]["queue"] == "test.queue"
        assert rabbitmq_span.stack
        assert isinstance(rabbitmq_span.stack, list)
        assert len(rabbitmq_span.stack) > 0

        cb.assert_called_once_with(self.obj, pika.spec.Basic.GetOk, properties, body)

    @mock.patch("pika.spec.Basic.Get")
    def test_basic_get_with_trace_context(self, _unused) -> None:
        self.obj._set_state(self.obj.OPEN)

        body = "Hello!"
        instana_headers = {
            "X-INSTANA-T": "0000000000000001",
            "X-INSTANA-S": "0000000000000002",
            "X-INSTANA-L": "1",
        }
        properties = pika.BasicProperties(headers=instana_headers)

        method_frame = pika.frame.Method(1, pika.spec.Basic.GetOk)
        header_frame = pika.frame.Header(1, len(body), properties)

        cb = mock.Mock()

        self.obj.basic_get("test.queue", cb)
        self.obj._on_getok(method_frame, header_frame, body)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # Trace context propagation
        assert rabbitmq_span.t == int(instana_headers["X-INSTANA-T"])
        assert rabbitmq_span.p == int(instana_headers["X-INSTANA-S"])

        # A new span has been started
        assert rabbitmq_span.s
        assert rabbitmq_span.p != rabbitmq_span.s

    @mock.patch("pika.spec.Basic.Consume")
    def test_basic_consume(self, _unused) -> None:
        self.obj._set_state(self.obj.OPEN)

        body = "Hello!"
        properties = pika.BasicProperties()

        method_frame = pika.frame.Method(
            1, pika.spec.Basic.Deliver(consumer_tag="test")
        )
        header_frame = pika.frame.Header(1, len(body), properties)

        cb = mock.Mock()

        self.obj.basic_consume("test.queue", cb, consumer_tag="test")
        self.obj._on_deliver(method_frame, header_frame, body)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # A new span has been started
        assert rabbitmq_span.t
        assert not rabbitmq_span.p
        assert rabbitmq_span.s

        # Error logging
        assert not rabbitmq_span.ec

        # Span tags
        assert not rabbitmq_span.data["rabbitmq"]["exchange"]
        assert rabbitmq_span.data["rabbitmq"]["sort"] == "consume"
        assert rabbitmq_span.data["rabbitmq"]["address"]
        assert rabbitmq_span.data["rabbitmq"]["queue"] == "test.queue"
        assert rabbitmq_span.stack
        assert isinstance(rabbitmq_span.stack, list)
        assert len(rabbitmq_span.stack) > 0

        cb.assert_called_once_with(self.obj, method_frame.method, properties, body)

    @mock.patch("pika.spec.Basic.Consume")
    def test_basic_consume_with_trace_context(self, _unused) -> None:
        self.obj._set_state(self.obj.OPEN)

        body = "Hello!"
        instana_headers = {
            "X-INSTANA-T": "0000000000000001",
            "X-INSTANA-S": "0000000000000002",
            "X-INSTANA-L": "1",
        }
        properties = pika.BasicProperties(headers=instana_headers)

        method_frame = pika.frame.Method(
            1, pika.spec.Basic.Deliver(consumer_tag="test")
        )
        header_frame = pika.frame.Header(1, len(body), properties)

        cb = mock.Mock()

        self.obj.basic_consume(
            queue="test.queue", on_message_callback=cb, consumer_tag="test"
        )
        self.obj._on_deliver(method_frame, header_frame, body)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        rabbitmq_span = spans[0]

        # Trace context propagation
        assert rabbitmq_span.t == int(instana_headers["X-INSTANA-T"])
        assert rabbitmq_span.p == int(instana_headers["X-INSTANA-S"])

        # A new span has been started
        assert rabbitmq_span.s
        assert rabbitmq_span.p != rabbitmq_span.s



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020


import contextvars
from typing import Any, Dict, Tuple
from instana.log import logger
from instana.propagators.format import Format
from instana.singletons import tracer
from instana.span.span import InstanaSpan
from instana.util.traceutils import get_tracer_tuple
from opentelemetry import trace, context

try:
    import celery
    from celery import registry, signals

    from urllib import parse

    client_token: Dict[str, Any] = {}
    worker_token: Dict[str, Any] = {}
    client_span = contextvars.ContextVar("client_span")
    worker_span = contextvars.ContextVar("worker_span")

    def _get_task_id(
        headers: Dict[str, Any],
        body: Tuple[str, Any],
    ) -> str:
        """
        Across Celery versions, the task id can exist in a couple of places.
        """
        id = headers.get("id", None)
        if id is None:
            id = body.get("id", None)
        return id

    def add_broker_attributes(
        span: InstanaSpan,
        broker_url: str,
    ) -> None:
        try:
            url = parse.urlparse(broker_url)

            # Add safety for edge case where scheme may not be a string
            url_scheme = str(url.scheme)
            span.set_attribute("scheme", url_scheme)

            span.set_attribute("host", url.hostname if url.hostname else "localhost")

            if not url.port:
                # Set default port if not specified
                if url_scheme == "redis":
                    span.set_attribute("port", "6379")
                elif "amqp" in url_scheme:
                    span.set_attribute("port", "5672")
                elif "sqs" in url_scheme:
                    span.set_attribute("port", "443")
            else:
                span.set_attribute("port", str(url.port))
        except Exception:
            logger.debug(f"Error parsing broker URL: {broker_url}", exc_info=True)

    @signals.task_prerun.connect
    def task_prerun(
        *args: Tuple[object, ...],
        **kwargs: Dict[str, Any],
    ) -> None:
        try:
            ctx = None

            task = kwargs.get("sender", None)
            task_id = kwargs.get("task_id", None)
            task = registry.tasks.get(task.name)

            headers = task.request.get("headers", {})
            if headers is not None:
                ctx = tracer.extract(
                    Format.HTTP_HEADERS, headers, disable_w3c_trace_context=True
                )

            span = tracer.start_span("celery-worker", span_context=ctx)
            span.set_attribute("task", task.name)
            span.set_attribute("task_id", task_id)
            add_broker_attributes(span, task.app.conf["broker_url"])

            ctx = trace.set_span_in_context(span)
            token = context.attach(ctx)
            worker_token["token"] = token
            worker_span.set(span)
        except Exception:
            logger.debug("celery-worker task_prerun: ", exc_info=True)

    @signals.task_postrun.connect
    def task_postrun(
        *args: Tuple[object, ...],
        **kwargs: Dict[str, Any],
    ) -> None:
        try:
            span = worker_span.get()

            if span.is_recording():
                span.end()
                worker_span.set(None)
            if "token" in worker_token:
                context.detach(worker_token.pop("token", None))
        except Exception:
            logger.debug("celery-worker after_task_publish: ", exc_info=True)

    @signals.task_failure.connect
    def task_failure(
        *args: Tuple[object, ...],
        **kwargs: Dict[str, Any],
    ) -> None:
        try:
            span = worker_span.get()
            if span.is_recording():
                span.set_attribute("success", False)
                exc = kwargs.get("exception", None)
                if exc:
                    span.record_exception(exc)
                else:
                    span.mark_as_errored()
        except Exception:
            logger.debug("celery-worker task_failure: ", exc_info=True)

    @signals.task_retry.connect
    def task_retry(
        *args: Tuple[object, ...],
        **kwargs: Dict[str, Any],
    ) -> None:
        try:
            span = worker_span.get()
            if span.is_recording():
                reason = kwargs.get("reason", None)
                if reason:
                    span.set_attribute("retry-reason", reason)
        except Exception:
            logger.debug("celery-worker task_failure: ", exc_info=True)

    @signals.before_task_publish.connect
    def before_task_publish(
        *args: Tuple[object, ...],
        **kwargs: Dict[str, Any],
    ) -> None:
        try:
            tracer, parent_span, _ = get_tracer_tuple()
            parent_context = parent_span.get_span_context() if parent_span else None

            if tracer:
                body = kwargs["body"]
                headers = kwargs["headers"]
                task_name = kwargs["sender"]
                task = registry.tasks.get(task_name)
                task_id = _get_task_id(headers, body)

                span = tracer.start_span("celery-client", span_context=parent_context)
                span.set_attribute("task", task_name)
                span.set_attribute("task_id", task_id)
                add_broker_attributes(span, task.app.conf["broker_url"])

                # Context propagation
                context_headers = {}
                tracer.inject(
                    span.context,
                    Format.HTTP_HEADERS,
                    context_headers,
                    disable_w3c_trace_context=True,
                )

                # Fix for broken header propagation
                # https://github.com/celery/celery/issues/4875
                task_headers = kwargs.get("headers") or {}
                task_headers.setdefault("headers", {})
                task_headers["headers"].update(context_headers)
                kwargs["headers"] = task_headers

                ctx = trace.set_span_in_context(span)
                token = context.attach(ctx)
                client_token["token"] = token
                client_span.set(span)
        except Exception:
            logger.debug("celery-client before_task_publish: ", exc_info=True)

    @signals.after_task_publish.connect
    def after_task_publish(
        *args: Tuple[object, ...],
        **kwargs: Dict[str, Any],
    ) -> None:
        try:
            span = client_span.get()
            if span.is_recording():
                span.end()
                client_span.set(None)
            if "token" in client_token:
                context.detach(client_token.pop("token", None))

        except Exception:
            logger.debug("celery-client after_task_publish: ", exc_info=True)

    logger.debug("Instrumenting celery")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

""" Collection helper for the Python runtime """
import gc
import importlib.metadata
import os
import platform
import resource
import sys
import threading
from types import ModuleType

from instana.collector.helpers.base import BaseHelper
from instana.log import logger
from instana.util import DictionaryOfStan
from instana.util.runtime import determine_service_name
from instana.version import VERSION

PATH_OF_DEPRECATED_INSTALLATION_VIA_HOST_AGENT = "/tmp/.instana/python"

PATH_OF_AUTOTRACE_WEBHOOK_SITEDIR = '/opt/instana/instrumentation/python/'

def is_autowrapt_instrumented():
    return 'instana' in os.environ.get('AUTOWRAPT_BOOTSTRAP', ())


def is_webhook_instrumented():
    return any(map(lambda p: PATH_OF_AUTOTRACE_WEBHOOK_SITEDIR in p, sys.path))


class RuntimeHelper(BaseHelper):
    """Helper class to collect snapshot and metrics for this Python runtime"""

    def __init__(self, collector):
        super(RuntimeHelper, self).__init__(collector)
        self.previous = DictionaryOfStan()
        self.previous_rusage = resource.getrusage(resource.RUSAGE_SELF)

        if gc.isenabled():
            self.previous_gc_count = gc.get_count()
        else:
            self.previous_gc_count = None

    def collect_metrics(self, **kwargs):
        plugin_data = dict()
        try:
            plugin_data["name"] = "com.instana.plugin.python"
            plugin_data["entityId"] = str(os.getpid())
            plugin_data["data"] = DictionaryOfStan()

            if hasattr(self.collector.agent, "announce_data"):
                try:
                    plugin_data["data"]["pid"] = self.collector.agent.announce_data.pid
                except Exception:
                    plugin_data["data"]["pid"] = str(os.getpid())
            else:
                plugin_data["data"]["pid"] = str(os.getpid())

            with_snapshot = kwargs.get("with_snapshot", False)
            self._collect_runtime_metrics(plugin_data, with_snapshot)

            if with_snapshot:
                self._collect_runtime_snapshot(plugin_data)
        except Exception:
            logger.debug("_collect_metrics: ", exc_info=True)
        return [plugin_data]

    def _collect_runtime_metrics(self, plugin_data, with_snapshot):
        if os.environ.get("INSTANA_DISABLE_METRICS_COLLECTION", False):
            return

        """ Collect up and return the runtime metrics """
        try:
            rusage = resource.getrusage(resource.RUSAGE_SELF)
            if gc.isenabled():
                self._collect_gc_metrics(plugin_data, with_snapshot)

            self._collect_thread_metrics(plugin_data, with_snapshot)

            value_diff = rusage.ru_utime - self.previous_rusage.ru_utime
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_utime",
                with_snapshot,
            )

            value_diff = rusage.ru_stime - self.previous_rusage.ru_stime
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_stime",
                with_snapshot,
            )

            self.apply_delta(
                rusage.ru_maxrss,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_maxrss",
                with_snapshot,
            )
            self.apply_delta(
                rusage.ru_ixrss,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_ixrss",
                with_snapshot,
            )
            self.apply_delta(
                rusage.ru_idrss,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_idrss",
                with_snapshot,
            )
            self.apply_delta(
                rusage.ru_isrss,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_isrss",
                with_snapshot,
            )

            value_diff = rusage.ru_minflt - self.previous_rusage.ru_minflt
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_minflt",
                with_snapshot,
            )

            value_diff = rusage.ru_majflt - self.previous_rusage.ru_majflt
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_majflt",
                with_snapshot,
            )

            value_diff = rusage.ru_nswap - self.previous_rusage.ru_nswap
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_nswap",
                with_snapshot,
            )

            value_diff = rusage.ru_inblock - self.previous_rusage.ru_inblock
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_inblock",
                with_snapshot,
            )

            value_diff = rusage.ru_oublock - self.previous_rusage.ru_oublock
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_oublock",
                with_snapshot,
            )

            value_diff = rusage.ru_msgsnd - self.previous_rusage.ru_msgsnd
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_msgsnd",
                with_snapshot,
            )

            value_diff = rusage.ru_msgrcv - self.previous_rusage.ru_msgrcv
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_msgrcv",
                with_snapshot,
            )

            value_diff = rusage.ru_nsignals - self.previous_rusage.ru_nsignals
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_nsignals",
                with_snapshot,
            )

            value_diff = rusage.ru_nvcsw - self.previous_rusage.ru_nvcsw
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_nvcsw",
                with_snapshot,
            )

            value_diff = rusage.ru_nivcsw - self.previous_rusage.ru_nivcsw
            self.apply_delta(
                value_diff,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "ru_nivcsw",
                with_snapshot,
            )
        except Exception:
            logger.debug("_collect_runtime_metrics", exc_info=True)
        finally:
            self.previous_rusage = rusage

    def _collect_gc_metrics(self, plugin_data, with_snapshot):
        try:
            gc_count = gc.get_count()
            gc_threshold = gc.get_threshold()

            self.apply_delta(
                gc_count[0],
                self.previous["data"]["metrics"]["gc"],
                plugin_data["data"]["metrics"]["gc"],
                "collect0",
                with_snapshot,
            )
            self.apply_delta(
                gc_count[1],
                self.previous["data"]["metrics"]["gc"],
                plugin_data["data"]["metrics"]["gc"],
                "collect1",
                with_snapshot,
            )
            self.apply_delta(
                gc_count[2],
                self.previous["data"]["metrics"]["gc"],
                plugin_data["data"]["metrics"]["gc"],
                "collect2",
                with_snapshot,
            )

            self.apply_delta(
                gc_threshold[0],
                self.previous["data"]["metrics"]["gc"],
                plugin_data["data"]["metrics"]["gc"],
                "threshold0",
                with_snapshot,
            )
            self.apply_delta(
                gc_threshold[1],
                self.previous["data"]["metrics"]["gc"],
                plugin_data["data"]["metrics"]["gc"],
                "threshold1",
                with_snapshot,
            )
            self.apply_delta(
                gc_threshold[2],
                self.previous["data"]["metrics"]["gc"],
                plugin_data["data"]["metrics"]["gc"],
                "threshold2",
                with_snapshot,
            )
        except Exception:
            logger.debug("_collect_gc_metrics", exc_info=True)

    def _collect_thread_metrics(self, plugin_data, with_snapshot):
        try:
            threads = threading.enumerate()
            daemon_threads = [thread.daemon is True for thread in threads].count(True)
            self.apply_delta(
                daemon_threads,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "daemon_threads",
                with_snapshot,
            )

            alive_threads = [thread.daemon is False for thread in threads].count(True)
            self.apply_delta(
                alive_threads,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "alive_threads",
                with_snapshot,
            )

            dummy_threads = [
                isinstance(thread, threading._DummyThread) for thread in threads
            ].count(True)  # pylint: disable=protected-access
            self.apply_delta(
                dummy_threads,
                self.previous["data"]["metrics"],
                plugin_data["data"]["metrics"],
                "dummy_threads",
                with_snapshot,
            )
        except Exception:
            logger.debug("_collect_thread_metrics", exc_info=True)

    def _collect_runtime_snapshot(self, plugin_data):
        """Gathers Python specific Snapshot information for this process"""
        snapshot_payload = {}
        try:
            snapshot_payload["name"] = determine_service_name()
            snapshot_payload["version"] = sys.version
            snapshot_payload["f"] = platform.python_implementation()  # flavor
            snapshot_payload["a"] = platform.architecture()[0]  # architecture
            snapshot_payload["versions"] = self.gather_python_packages()
            snapshot_payload["iv"] = VERSION

            if is_autowrapt_instrumented():
                snapshot_payload['m'] = 'Autowrapt'
            elif is_webhook_instrumented():
                snapshot_payload['m'] = 'AutoTrace'
            else:
                snapshot_payload["m"] = "Manual"

            try:
                from django.conf import (
                    settings,  # pylint: disable=import-outside-toplevel
                )

                if hasattr(settings, "MIDDLEWARE") and settings.MIDDLEWARE is not None:
                    snapshot_payload["djmw"] = settings.MIDDLEWARE
                elif (
                    hasattr(settings, "MIDDLEWARE_CLASSES")
                    and settings.MIDDLEWARE_CLASSES is not None
                ):
                    snapshot_payload["djmw"] = settings.MIDDLEWARE_CLASSES
            except Exception:
                pass
        except Exception:
            logger.debug("collect_snapshot: ", exc_info=True)

        plugin_data["data"]["snapshot"] = snapshot_payload

    def gather_python_packages(self):
        """Collect up the list of modules in use"""
        if os.environ.get("INSTANA_DISABLE_PYTHON_PACKAGE_COLLECTION"):
            return {"instana": VERSION}

        versions = {}
        try:
            sys_packages = sys.modules.copy()

            for pkg_name in sys_packages:
                # Don't report submodules (e.g. django.x, django.y, django.z)
                # Skip modules that begin with underscore
                if ("." in pkg_name) or pkg_name[0] == "_":
                    continue

                # Skip builtins
                if pkg_name in ["sys", "curses"]:
                    continue

                if sys_packages[pkg_name]:
                    try:
                        pkg_info = sys_packages[pkg_name].__dict__
                        if "__version__" in pkg_info:
                            if isinstance(pkg_info["__version__"], str):
                                versions[pkg_name] = pkg_info["__version__"]
                            else:
                                versions[pkg_name] = self.jsonable(
                                    pkg_info["__version__"]
                                )
                        elif "version" in pkg_info:
                            versions[pkg_name] = self.jsonable(pkg_info["version"])
                        else:
                            versions[pkg_name] = importlib.metadata.version(pkg_name)
                    except importlib.metadata.PackageNotFoundError:
                        pass
                    except Exception:
                        logger.debug(
                            "gather_python_packages: could not process module: %s",
                            pkg_name,
                        )

            # Manually set our package version
            versions["instana"] = VERSION
        except Exception:
            logger.debug("gather_python_packages", exc_info=True)

        return versions

    def jsonable(self, value):
        try:
            if callable(value):
                try:
                    result = value()
                except Exception:
                    result = "Unknown"
            elif isinstance(value, ModuleType):
                result = value
            else:
                result = value
            return str(result)
        except Exception:
            logger.debug("jsonable: ", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2017

"""
This module contains the classes that represents spans.

InstanaSpan - the OpenTelemetry based span used during tracing

When an InstanaSpan is finished, it is converted into either an SDKSpan
or RegisteredSpan depending on type.

BaseSpan: Base class containing the commonalities for the two descendants
  - SDKSpan: Class that represents an SDK type span
  - RegisteredSpan: Class that represents a Registered type span
"""

from threading import Lock
from time import time_ns
from typing import Dict, Optional, Sequence, Union

from opentelemetry.context import get_value
from opentelemetry.context.context import Context
from opentelemetry.trace import (
    _SPAN_KEY,
    DEFAULT_TRACE_OPTIONS,
    DEFAULT_TRACE_STATE,
    INVALID_SPAN_ID,
    INVALID_TRACE_ID,
    Span,
)
from opentelemetry.trace.span import NonRecordingSpan
from opentelemetry.trace.status import Status, StatusCode
from opentelemetry.util import types

from instana.log import logger
from instana.recorder import StanRecorder
from instana.span.kind import HTTP_SPANS
from instana.span.readable_span import Event, ReadableSpan
from instana.span_context import SpanContext


class InstanaSpan(Span, ReadableSpan):
    def __init__(
        self,
        name: str,
        context: SpanContext,
        span_processor: StanRecorder,
        parent_id: Optional[str] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        attributes: types.Attributes = {},
        events: Sequence[Event] = [],
        status: Optional[Status] = Status(StatusCode.UNSET),
    ) -> None:
        super().__init__(
            name=name,
            context=context,
            parent_id=parent_id,
            start_time=start_time,
            end_time=end_time,
            attributes=attributes,
            events=events,
            status=status,
            # kind=kind,
        )
        self._span_processor = span_processor
        self._lock = Lock()

    def get_span_context(self) -> SpanContext:
        return self._context

    def set_attributes(self, attributes: Dict[str, types.AttributeValue]) -> None:
        if not self._attributes:
            self._attributes = {}

        with self._lock:
            for key, value in attributes.items():
                self._attributes[key] = value

    def set_attribute(self, key: str, value: types.AttributeValue) -> None:
        return self.set_attributes({key: value})

    def update_name(self, name: str) -> None:
        with self._lock:
            self._name = name

    def is_recording(self) -> bool:
        return self._end_time is None

    def set_status(
        self,
        status: Union[Status, StatusCode],
        description: Optional[str] = None,
    ) -> None:
        # Ignore future calls if status is already set to OK
        # Ignore calls to set to StatusCode.UNSET
        if isinstance(status, Status):
            if (
                self._status
                and self._status.status_code is StatusCode.OK
                or status.status_code is StatusCode.UNSET
            ):
                return
            if description is not None:
                logger.warning(
                    "Description %s ignored. Use either `Status` or `(StatusCode, Description)`",
                    description,
                )
            self._status = status
        elif isinstance(status, StatusCode):
            if (
                self._status
                and self._status.status_code is StatusCode.OK
                or status is StatusCode.UNSET
            ):
                return
            self._status = Status(status, description)

    def add_event(
        self,
        name: str,
        attributes: types.Attributes = None,
        timestamp: Optional[int] = None,
    ) -> None:
        event = Event(
            name=name,
            attributes=attributes,
            timestamp=timestamp,
        )

        self._events.append(event)

    def record_exception(
        self,
        exception: Exception,
        attributes: types.Attributes = None,
        timestamp: Optional[int] = None,
        escaped: bool = False,
    ) -> None:
        """
        Records an exception as a span event. This will record pertinent info from the exception and
        assure that this span is marked as errored.
        """
        try:
            message = ""
            self.mark_as_errored()
            if hasattr(exception, "__str__") and len(str(exception)) > 0:
                message = str(exception)
            elif hasattr(exception, "message") and exception.message is not None:
                message = exception.message
            else:
                message = repr(exception)

            if self.name in ["rpc-server", "rpc-client"]:
                self.set_attribute("rpc.error", message)
            elif self.name == "mysql":
                self.set_attribute("mysql.error", message)
            elif self.name == "postgres":
                self.set_attribute("pg.error", message)
            elif self.name in HTTP_SPANS:
                self.set_attribute("http.error", message)
            elif self.name in ["celery-client", "celery-worker"]:
                self.set_attribute("error", message)
            elif self.name == "sqlalchemy":
                self.set_attribute("sqlalchemy.err", message)
            elif self.name == "aws.lambda.entry":
                self.set_attribute("lambda.error", message)
            else:
                _attributes = {"message": message}
                if attributes:
                    _attributes.update(attributes)
                self.add_event(
                    name="exception", attributes=_attributes, timestamp=timestamp
                )
        except Exception:
            logger.debug("span.record_exception", exc_info=True)
            raise

    def _readable_span(self) -> ReadableSpan:
        return ReadableSpan(
            name=self.name,
            context=self.context,
            parent_id=self.parent_id,
            start_time=self.start_time,
            end_time=self.end_time,
            attributes=self.attributes,
            events=self.events,
            status=self.status,
            stack=self.stack,
            # kind=self.kind,
        )

    def end(self, end_time: Optional[int] = None) -> None:
        with self._lock:
            self._end_time = end_time if end_time else time_ns()
            self._duration = self._end_time - self._start_time

        self._span_processor.record_span(self._readable_span())

    def mark_as_errored(self, attributes: types.Attributes = None) -> None:
        """
        Mark this span as errored.

        @param attributes: optional attributes to add to the span
        """
        try:
            ec = self.attributes.get("ec", 0)
            self.set_attribute("ec", ec + 1)

            if attributes is not None and isinstance(attributes, dict):
                for key in attributes:
                    self.set_attribute(key, attributes[key])
        except Exception:
            logger.debug("span.mark_as_errored", exc_info=True)

    def assure_errored(self) -> None:
        """
        Make sure that this span is marked as errored.
        @return: None
        """
        try:
            ec = self.attributes.get("ec", None)
            if ec is None or ec == 0:
                self.set_attribute("ec", 1)
        except Exception:
            logger.debug("span.assure_errored", exc_info=True)


INVALID_SPAN_CONTEXT = SpanContext(
    trace_id=INVALID_TRACE_ID,
    span_id=INVALID_SPAN_ID,
    is_remote=False,
    trace_flags=DEFAULT_TRACE_OPTIONS,
    trace_state=DEFAULT_TRACE_STATE,
)
INVALID_SPAN = NonRecordingSpan(INVALID_SPAN_CONTEXT)


def get_current_span(context: Optional[Context] = None) -> InstanaSpan:
    """Retrieve the current span.

    Args:
        context: A Context object. If one is not passed, the
            default current context is used instead.

    Returns:
        The Span set in the context if it exists. INVALID_SPAN otherwise.
    """
    span = get_value(_SPAN_KEY, context=context)
    if span is None or not isinstance(span, InstanaSpan):
        return INVALID_SPAN
    return span



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import logging

import instana

from flask import Flask, request
from google.cloud import pubsub_v1

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

app = Flask(__name__)
app.debug = True
app.use_reloader = True

# :Development:
# Use PubSub Emulator exposed at :8432 for local testing and uncomment below
# os.environ["PUBSUB_EMULATOR_HOST"] = "localhost:8432"

PROJECT_ID = 'k8s-brewery'
TOPIC_NAME = 'python-test-topic'
SUBSCRIPTION_ID = 'python-test-subscription'

publisher = pubsub_v1.PublisherClient()
subscriber = pubsub_v1.SubscriberClient()

TOPIC_PATH = publisher.topic_path(PROJECT_ID, TOPIC_NAME)
SUBSCRIPTION_PATH = subscriber.subscription_path(PROJECT_ID, SUBSCRIPTION_ID)


@app.route('/')
def home():
    return "Welcome to PubSub testing."


@app.route('/create')
def create_topic():
    """
    Usage: /create?topic=<your-topic-name-here>
    """
    topic = request.args.get('topic')
    print(topic, type(topic))

    try:
        publisher.create_topic(TOPIC_PATH)
        return "Topic Created"
    except Exception as e:
        return "Topic Creation Failed: %s" % e


@app.route('/publish')
def publish():
    """
    Usage: /publish?message=<your-message-here>
    """
    msg = request.args.get('message').encode('utf-8')
    publisher.publish(TOPIC_PATH, msg, origin='instana-test')
    return "Published msg: %s" % msg


@app.route('/consume')
def consume():
    """
    Usage: /consume
    * Run it in a different browser tab. Logs on terminal.
    """

    # Async
    def callback_handler(message):
        print('MESSAGE: ', message, type(message))
        print(message.data)
        message.ack()

    future = subscriber.subscribe(SUBSCRIPTION_PATH, callback_handler)

    try:
        res = future.result()
        print('CALLBACK: ', res, type(res))
    except KeyboardInterrupt:
        future.cancel()
    return "Consumer closed."


if __name__ == '__main__':
    app.run(host='127.0.0.1', port='10811')



# (c) Copyright IBM Corp. 2023

"""
Collector for EKS Pods on AWS Fargate: Manages the periodic collection of metrics & snapshot data
"""

from time import time

from instana.collector.base import BaseCollector
from instana.collector.helpers.eks.process import EKSFargateProcessHelper
from instana.collector.helpers.runtime import RuntimeHelper
from instana.collector.utils import format_span
from instana.log import logger
from instana.util import DictionaryOfStan


class EKSFargateCollector(BaseCollector):
    """Collector for EKS Pods on AWS Fargate"""

    def __init__(self, agent):
        super(EKSFargateCollector, self).__init__(agent)
        logger.debug("Loading Collector for EKS Pods on AWS Fargate ")

        self.snapshot_data = DictionaryOfStan()
        self.snapshot_data_sent = False
        self.podname = agent.podname
        self.helpers.append(EKSFargateProcessHelper(self))
        self.helpers.append(RuntimeHelper(self))

    def should_send_snapshot_data(self):
        return int(time()) - self.snapshot_data_last_sent > self.snapshot_data_interval

    def prepare_payload(self):
        payload = DictionaryOfStan()
        payload["spans"] = []
        payload["metrics"]["plugins"] = []

        try:
            if not self.span_queue.empty():
                payload["spans"] = format_span(self.queued_spans())

            with_snapshot = self.should_send_snapshot_data()

            plugins = []
            for helper in self.helpers:
                plugins.extend(helper.collect_metrics(with_snapshot=with_snapshot))

            payload["metrics"]["plugins"] = plugins

            if with_snapshot:
                self.snapshot_data_last_sent = int(time())
        except Exception:
            logger.debug("prepare_payload error", exc_info=True)

        return payload



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020


import os

from typing import Any, Optional, TypeVar, Dict, List, Tuple

from instana.log import logger
from instana.span_context import SpanContext
from instana.util.ids import header_to_id, header_to_long_id, hex_id, internal_id, internal_id_limited, hex_id_limited
from instana.w3c_trace_context.traceparent import Traceparent
from instana.w3c_trace_context.tracestate import Tracestate

from opentelemetry.trace import (
    INVALID_SPAN_ID,
    INVALID_TRACE_ID,
)

# The carrier, typed here as CarrierT, can be a dict, a list, or a tuple.
# Using the trace header as an example, it can be in the following forms
# for extraction:
#   X-Instana-T
#   HTTP_X_INSTANA_T
#
# The second form above is found in places like Django middleware for
# incoming requests.
#
# For injection, we only support the standard format:
#   X-Instana-T
CarrierT = TypeVar("CarrierT", Dict, List, Tuple)


class BasePropagator(object):
    HEADER_KEY_T = 'X-INSTANA-T'
    HEADER_KEY_S = 'X-INSTANA-S'
    HEADER_KEY_L = 'X-INSTANA-L'
    HEADER_KEY_SYNTHETIC = 'X-INSTANA-SYNTHETIC'
    HEADER_KEY_TRACEPARENT = "traceparent"
    HEADER_KEY_TRACESTATE = "tracestate"
    HEADER_KEY_SERVER_TIMING = "Server-Timing"

    LC_HEADER_KEY_T = 'x-instana-t'
    LC_HEADER_KEY_S = 'x-instana-s'
    LC_HEADER_KEY_L = 'x-instana-l'
    LC_HEADER_KEY_SYNTHETIC = 'x-instana-synthetic'
    LC_HEADER_KEY_SERVER_TIMING = "server-timing"

    ALT_LC_HEADER_KEY_T = 'http_x_instana_t'
    ALT_LC_HEADER_KEY_S = 'http_x_instana_s'
    ALT_LC_HEADER_KEY_L = 'http_x_instana_l'
    ALT_LC_HEADER_KEY_SYNTHETIC = 'http_x_instana_synthetic'
    ALT_HEADER_KEY_TRACEPARENT = "http_traceparent"
    ALT_HEADER_KEY_TRACESTATE = "http_tracestate"
    ALT_LC_HEADER_KEY_SERVER_TIMING = "http_server_timing"

    # ByteArray variations
    B_HEADER_KEY_T = b'x-instana-t'
    B_HEADER_KEY_S = b'x-instana-s'
    B_HEADER_KEY_L = b'x-instana-l'
    B_HEADER_KEY_SYNTHETIC = b'x-instana-synthetic'
    B_HEADER_KEY_TRACEPARENT = b'traceparent'
    B_HEADER_KEY_TRACESTATE = b'tracestate'
    B_HEADER_KEY_SERVER_TIMING = b"server-timing"

    B_ALT_LC_HEADER_KEY_T = b'http_x_instana_t'
    B_ALT_LC_HEADER_KEY_S = b'http_x_instana_s'
    B_ALT_LC_HEADER_KEY_L = b'http_x_instana_l'
    B_ALT_LC_HEADER_KEY_SYNTHETIC = b'http_x_instana_synthetic'
    B_ALT_HEADER_KEY_TRACEPARENT = b'http_traceparent'
    B_ALT_HEADER_KEY_TRACESTATE = b'http_tracestate'
    B_ALT_LC_HEADER_KEY_SERVER_TIMING = b"http_server_timing"

    def __init__(self):
        self._tp = Traceparent()
        self._ts = Tracestate()

    @staticmethod
    def extract_headers_dict(carrier: CarrierT) -> Optional[Dict]:
        """
        This method converts the incoming carrier into a dict.

        :param carrier: CarrierT
        :return: Dict | None
        """
        dc = None
        try:
            if isinstance(carrier, dict):
                dc = carrier
            elif hasattr(carrier, "__dict__"):
                dc = carrier.__dict__
            else:
                dc = dict(carrier)
        except Exception:
            logger.debug(f"base_propagator extract_headers_dict: Couldn't convert - {carrier}")

        return dc

    @staticmethod
    def _get_ctx_level(level: str) -> int:
        """
        Extract the level value and return it, as it may include correlation values.

        :param level: str
        :return: int
        """
        try:
            ctx_level = int(level.split(",")[0]) if level else 1
        except Exception:
            ctx_level = 1
        return ctx_level

    @staticmethod
    def _get_correlation_properties(level:str):
        """
        Get the correlation values if they are present.

        :param level: str
        :return: Tuple[Any, Any] - correlation_type, correlation_id
        """
        correlation_type, correlation_id = [None] * 2
        try:
            correlation_type = level.split(",")[1].split("correlationType=")[1].split(";")[0]
            if "correlationId" in level:
                correlation_id = level.split(",")[1].split("correlationId=")[1].split(";")[0]
        except Exception:
            logger.debug("extract instana correlation type/id error:", exc_info=True)
        
        return correlation_type, correlation_id

    def _get_participating_trace_context(self, span_context: SpanContext):
        """
        This method is called for getting the updated traceparent and tracestate values.

        :param span_context: SpanContext
        :return: traceparent, tracestate
        """
        if span_context.long_trace_id and not span_context.trace_parent:
            tp_trace_id = span_context.long_trace_id
        else:
            tp_trace_id = span_context.trace_id
        traceparent = span_context.traceparent
        tracestate = span_context.tracestate
        traceparent = self._tp.update_traceparent(traceparent, tp_trace_id, span_context.span_id, span_context.level)

        # In suppression mode do not update the tracestate and
        # do not add the 'in=' key-value pair to the incoming tracestate
        # Just propagate the incoming tracestate (if any) unchanged.
        if span_context.suppression:
            return traceparent, tracestate

        tracestate = self._ts.update_tracestate(tracestate, hex_id_limited(span_context.trace_id), hex_id(span_context.span_id))
        return traceparent, tracestate

    def __determine_span_context(
            self, 
            trace_id: int, 
            span_id: int, 
            level: str, 
            synthetic: bool, 
            traceparent, 
            tracestate,
            disable_w3c_trace_context: bool,
        ) -> SpanContext:
        """
        This method determines the span context depending on a set of conditions being met
        Detailed description of the conditions can be found in the instana internal technical-documentation,
        under section http-processing-for-instana-tracers.

        :param trace_id: int - instana trace id
        :param span_id: int - instana span id
        :param level: str - instana level
        :param synthetic: bool - instana synthetic
        :param traceparent:
        :param tracestate:
        :param disable_w3c_trace_context: bool - flag used to enable w3c trace context only on HTTP requests
        :return: SpanContext
        """
        correlation = False
        disable_traceparent = os.environ.get("INSTANA_DISABLE_W3C_TRACE_CORRELATION", "")
        instana_ancestor = None

        if level and "correlationType" in level:
            trace_id, span_id = [None] * 2
            correlation = True

        (
            ctx_level,
            ctx_synthetic,
            ctx_trace_parent, 
            ctx_instana_ancestor,
            ctx_long_trace_id,
            ctx_correlation_type,
            ctx_correlation_id,
            ctx_traceparent,
            ctx_tracestate,
        ) = [None] * 9

        ctx_level = self._get_ctx_level(level)
        ctx_trace_id = trace_id if ctx_level > 0 else None
        ctx_span_id = span_id if ctx_level > 0 else None

        if (
            trace_id
            and span_id
            and trace_id != INVALID_TRACE_ID
            and span_id != INVALID_SPAN_ID
        ):
            ctx_synthetic = synthetic

            hex_trace_id = hex_id(trace_id)
            if len(hex_trace_id) > 16:
                ctx_long_trace_id = hex_trace_id

        elif not disable_w3c_trace_context and traceparent and not trace_id and not span_id:
            _, tp_trace_id, tp_parent_id, _ = self._tp.get_traceparent_fields(traceparent)

            if tracestate and "in=" in tracestate:
                instana_ancestor = self._ts.get_instana_ancestor(tracestate)

            if disable_traceparent == "":
                ctx_trace_id = hex_id_limited(tp_trace_id)
                ctx_span_id = tp_parent_id
                ctx_synthetic = synthetic
                ctx_trace_parent = True
                ctx_instana_ancestor = instana_ancestor
                ctx_long_trace_id = tp_trace_id
            else:
                if instana_ancestor:
                    ctx_trace_id = instana_ancestor.t
                    ctx_span_id = instana_ancestor.p
                    ctx_synthetic = synthetic

        elif synthetic:
            ctx_synthetic = synthetic

        if correlation:
            ctx_correlation_type, ctx_correlation_id = self._get_correlation_properties(level)

        if traceparent:
            ctx_traceparent = traceparent
            ctx_tracestate = tracestate

        if ctx_trace_id:
            if isinstance(ctx_trace_id, int):
                # check if ctx_trace_id is a valid internal trace id
                if (ctx_trace_id <=  2**64 - 1):
                    trace_id = ctx_trace_id
                else:
                    trace_id = internal_id(hex_id_limited(ctx_trace_id))
            else:
                trace_id = internal_id(ctx_trace_id)
        else:
            trace_id = INVALID_TRACE_ID

        return SpanContext(
            trace_id=trace_id,
            span_id=internal_id_limited(ctx_span_id) if ctx_span_id else INVALID_SPAN_ID,
            is_remote=False,
            level=ctx_level,
            synthetic=ctx_synthetic,
            trace_parent=ctx_trace_parent,
            instana_ancestor=ctx_instana_ancestor,
            long_trace_id=ctx_long_trace_id,
            correlation_type=ctx_correlation_type,
            correlation_id=ctx_correlation_id,
            traceparent=ctx_traceparent,
            tracestate=ctx_tracestate,
        )


    def extract_instana_headers(self, dc: Dict[str, Any]) -> Tuple[Optional[int], Optional[int], Optional[str], Optional[bool]]:
        """
        Search carrier for the *HEADER* keys and return the tracing key-values.

        :param dc: Dict - The dict potentially containing context
        :return: Tuple[Optional[int], Optional[int], Optional[str], Optional[bool]] - trace_id, span_id, level, synthetic
        """
        trace_id, span_id, level, synthetic = [None] * 4

        # Headers can exist in the standard X-Instana-T/S format or the alternate HTTP_X_INSTANA_T/S style
        try:
            trace_id = dc.get(self.LC_HEADER_KEY_T) or dc.get(self.ALT_LC_HEADER_KEY_T) or dc.get(
                self.B_HEADER_KEY_T) or dc.get(self.B_ALT_LC_HEADER_KEY_T)
            if trace_id:
                trace_id = header_to_long_id(trace_id)

            span_id = dc.get(self.LC_HEADER_KEY_S) or dc.get(self.ALT_LC_HEADER_KEY_S) or dc.get(
                self.B_HEADER_KEY_S) or dc.get(self.B_ALT_LC_HEADER_KEY_S)
            if span_id:
                span_id = header_to_id(span_id)

            level = dc.get(self.LC_HEADER_KEY_L) or dc.get(self.ALT_LC_HEADER_KEY_L) or dc.get(
                self.B_HEADER_KEY_L) or dc.get(self.B_ALT_LC_HEADER_KEY_L)
            if level and isinstance(level, bytes):
                level = level.decode("utf-8")

            synthetic = dc.get(self.LC_HEADER_KEY_SYNTHETIC) or dc.get(self.ALT_LC_HEADER_KEY_SYNTHETIC) or dc.get(
                self.B_HEADER_KEY_SYNTHETIC) or dc.get(self.B_ALT_LC_HEADER_KEY_SYNTHETIC)
            if synthetic:
                synthetic = synthetic in ['1', b'1']

        except Exception:
            logger.debug("extract error:", exc_info=True)

        return trace_id, span_id, level, synthetic

    def __extract_w3c_trace_context_headers(self, dc):
        """
        Search carrier for the *HEADER* keys and return the tracing key-values

        :param dc: The dict or list potentially containing context
        :return: traceparent, tracestate
        """
        traceparent, tracestate = [None] * 2

        try:
            traceparent = dc.get(self.HEADER_KEY_TRACEPARENT) or dc.get(self.ALT_HEADER_KEY_TRACEPARENT) or dc.get(
                self.B_HEADER_KEY_TRACEPARENT) or dc.get(self.B_ALT_HEADER_KEY_TRACEPARENT)
            if traceparent and isinstance(traceparent, bytes):
                traceparent = traceparent.decode("utf-8")

            tracestate = dc.get(self.HEADER_KEY_TRACESTATE) or dc.get(self.ALT_HEADER_KEY_TRACESTATE) or dc.get(
                self.B_HEADER_KEY_TRACESTATE) or dc.get(self.B_ALT_HEADER_KEY_TRACESTATE)
            if tracestate and isinstance(tracestate, bytes):
                tracestate = tracestate.decode("utf-8")

        except Exception:
            logger.debug("extract error:", exc_info=True)

        return traceparent, tracestate

    def extract(self, carrier: CarrierT, disable_w3c_trace_context: bool = False) -> Optional[SpanContext]:
        """
        This method overrides one of the Base classes as with the introduction 
        of W3C trace context for the HTTP requests more extracting steps and 
        logic was required.

        :param disable_w3c_trace_context:
        :param carrier:
        :return: the context or None
        """
        try:
            traceparent, tracestate = [None] * 2
            headers = self.extract_headers_dict(carrier=carrier)
            if headers is None:
                return None
            headers = {k.lower(): v for k, v in headers.items()}

            trace_id, span_id, level, synthetic = self.extract_instana_headers(dc=headers)
            if not disable_w3c_trace_context:
                traceparent, tracestate = self.__extract_w3c_trace_context_headers(dc=headers)

            if traceparent:
                traceparent = self._tp.validate(traceparent)

            if trace_id is None:
                trace_id = INVALID_TRACE_ID
            if span_id is None:
                span_id = INVALID_SPAN_ID

            span_context = self.__determine_span_context(
                trace_id,
                span_id,
                level,
                synthetic,
                traceparent,
                tracestate,
                disable_w3c_trace_context,
            )
            return span_context

        except Exception:
            logger.debug("base_propagator extract error:", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import asyncio
from typing import Generator

import aiohttp
import pytest

from instana.singletons import agent, tracer
from instana.util.ids import hex_id
from tests.helpers import testenv


class TestAiohttpServer:
    async def fetch(self, session, url, headers=None, params=None):
        try:
            async with session.get(url, headers=headers, params=params) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Load test server application
        import tests.apps.aiohttp_app  # noqa: F401

        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # New event loop for every test
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        yield

    def test_server_get(self):
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["aiohttp_server"] + "/")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aioclient_span.t == traceId
        assert aioserver_span.t == traceId

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s

        # Synthetic
        assert not test_span.sy
        assert not aioclient_span.sy
        assert not aioserver_span.sy

        # Error logging
        assert not test_span.ec
        assert not aioclient_span.ec
        assert not aioserver_span.ec

        assert aioserver_span.n == "aiohttp-server"
        assert aioserver_span.data["http"]["status"] == 200
        assert aioserver_span.data["http"]["url"] == f"{testenv['aiohttp_server']}/"
        assert aioserver_span.data["http"]["method"] == "GET"
        assert not aioserver_span.stack

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(aioserver_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_server_get_204(self):
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["aiohttp_server"] + "/204")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        trace_id = test_span.t
        assert aioclient_span.t == trace_id
        assert aioserver_span.t == trace_id

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s

        # Synthetic
        assert not test_span.sy
        assert not aioclient_span.sy
        assert not aioserver_span.sy

        # Error logging
        assert not test_span.ec
        assert not aioclient_span.ec
        assert not aioserver_span.ec

        assert aioserver_span.n == "aiohttp-server"
        assert aioserver_span.data["http"]["status"] == 204
        assert aioserver_span.data["http"]["url"] == f"{testenv['aiohttp_server']}/204"
        assert aioserver_span.data["http"]["method"] == "GET"
        assert not aioserver_span.stack

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(trace_id)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(aioserver_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(trace_id)}"

    def test_server_synthetic_request(self):
        async def test():
            headers = {"X-INSTANA-SYNTHETIC": "1"}

            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(
                        session, testenv["aiohttp_server"] + "/", headers=headers
                    )

        response = self.loop.run_until_complete(test())
        assert response

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        assert aioserver_span.sy
        assert not aioclient_span.sy
        assert not test_span.sy

    def test_server_get_with_params_to_scrub(self):
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(
                        session,
                        testenv["aiohttp_server"],
                        params={"secret": "iloveyou"},
                    )

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aioclient_span.t == traceId
        assert aioserver_span.t == traceId

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s

        # Error logging
        assert not test_span.ec
        assert not aioclient_span.ec
        assert not aioserver_span.ec

        assert aioserver_span.n == "aiohttp-server"
        assert aioserver_span.data["http"]["status"] == 200
        assert aioserver_span.data["http"]["url"] == f"{testenv['aiohttp_server']}/"
        assert aioserver_span.data["http"]["method"] == "GET"
        assert aioserver_span.data["http"]["params"] == "secret=<redacted>"
        assert not aioserver_span.stack

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(aioserver_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_server_request_header_capture(self):
        original_extra_http_headers = agent.options.extra_http_headers

        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    # Hack together a manual custom headers list
                    agent.options.extra_http_headers = [
                        "X-Capture-This",
                        "X-Capture-That",
                    ]

                    headers = dict()
                    headers["X-Capture-This"] = "this"
                    headers["X-Capture-That"] = "that"

                    return await self.fetch(
                        session,
                        testenv["aiohttp_server"],
                        headers=headers,
                        params={"secret": "iloveyou"},
                    )

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aioclient_span.t == traceId
        assert aioserver_span.t == traceId

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s

        # Error logging
        assert not test_span.ec
        assert not aioclient_span.ec
        assert not aioserver_span.ec

        assert aioserver_span.n == "aiohttp-server"
        assert aioserver_span.data["http"]["status"] == 200
        assert aioserver_span.data["http"]["url"] == f"{testenv['aiohttp_server']}/"
        assert aioserver_span.data["http"]["method"] == "GET"
        assert aioserver_span.data["http"]["params"] == "secret=<redacted>"
        assert not aioserver_span.stack

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(aioserver_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

        assert "X-Capture-This" in aioserver_span.data["http"]["header"]
        assert aioserver_span.data["http"]["header"]["X-Capture-This"] == "this"
        assert "X-Capture-That" in aioserver_span.data["http"]["header"]
        assert aioserver_span.data["http"]["header"]["X-Capture-That"] == "that"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_server_response_header_capture(self):
        original_extra_http_headers = agent.options.extra_http_headers

        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    # Hack together a manual custom headers list
                    agent.options.extra_http_headers = [
                        "X-Capture-This-Too",
                        "X-Capture-That-Too",
                    ]

                    return await self.fetch(
                        session,
                        testenv["aiohttp_server"] + "/response_headers"
                    )

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aioclient_span.t == traceId
        assert aioserver_span.t == traceId

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s

        # Error logging
        assert not test_span.ec
        assert not aioclient_span.ec
        assert not aioserver_span.ec

        assert aioserver_span.n == "aiohttp-server"
        assert aioserver_span.data["http"]["status"] == 200
        assert aioserver_span.data["http"]["url"] == f"{testenv['aiohttp_server']}/response_headers"
        assert aioserver_span.data["http"]["method"] == "GET"
        assert not aioserver_span.stack

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(aioserver_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

        assert "X-Capture-This-Too" in aioserver_span.data["http"]["header"]
        assert aioserver_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in aioserver_span.data["http"]["header"]
        assert aioserver_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_server_get_401(self):
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["aiohttp_server"] + "/401")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aioclient_span.t == traceId
        assert aioserver_span.t == traceId

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s

        # Error logging
        assert not test_span.ec
        assert not aioclient_span.ec
        assert not aioserver_span.ec

        assert aioserver_span.n == "aiohttp-server"
        assert aioserver_span.data["http"]["status"] == 401
        assert aioserver_span.data["http"]["url"] == f"{testenv['aiohttp_server']}/401"
        assert aioserver_span.data["http"]["method"] == "GET"
        assert not aioserver_span.stack

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(aioserver_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_server_get_500(self):
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["aiohttp_server"] + "/500")

        response = self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aioclient_span.t == traceId
        assert aioserver_span.t == traceId

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s

        # Error logging
        assert not test_span.ec
        assert aioclient_span.ec == 1
        assert aioserver_span.ec == 1

        assert aioserver_span.n == "aiohttp-server"
        assert aioserver_span.data["http"]["status"] == 500
        assert aioserver_span.data["http"]["url"] == f"{testenv['aiohttp_server']}/500"
        assert aioserver_span.data["http"]["method"] == "GET"
        assert not aioserver_span.stack

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(traceId)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(aioserver_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(traceId)}"

    def test_server_get_exception(self):
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(
                        session, testenv["aiohttp_server"] + "/exception"
                    )

        response = self.loop.run_until_complete(test())
        assert response

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aioclient_span.t == traceId
        assert aioserver_span.t == traceId

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s

        # Error logging
        assert not test_span.ec
        assert aioclient_span.ec == 1
        assert aioserver_span.ec == 1

        assert aioserver_span.n == "aiohttp-server"
        assert aioserver_span.data["http"]["status"] == 500
        assert (
            aioserver_span.data["http"]["url"]
            == f"{testenv['aiohttp_server']}/exception"
        )
        assert aioserver_span.data["http"]["method"] == "GET"
        assert not aioserver_span.stack

        assert aioclient_span.n == "aiohttp-client"
        assert aioclient_span.data["http"]["status"] == 500
        assert aioclient_span.data["http"]["error"] == "Internal Server Error"
        assert aioclient_span.stack
        assert isinstance(aioclient_span.stack, list)
        assert len(aioclient_span.stack) > 1


class TestAiohttpServerMiddleware:
    async def fetch(self, session, url, headers=None, params=None):
        try:
            async with session.get(url, headers=headers, params=params) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Load test server application
        import tests.apps.aiohttp_app2  # noqa: F401

        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # New event loop for every test
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        yield

    def test_server_get(self):
        async def test():
            with tracer.start_as_current_span("test"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["aiohttp_server"] + "/")

        response = self.loop.run_until_complete(test())
        assert response

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        aioserver_span = spans[0]
        aioclient_span = spans[1]
        test_span = spans[2]

        # Same traceId
        traceId = test_span.t
        assert aioclient_span.t == traceId
        assert aioserver_span.t == traceId

        # Parent relationships
        assert aioclient_span.p == test_span.s
        assert aioserver_span.p == aioclient_span.s
